@@ -15,10 +15,10 @@ from pyhanko.pdf_utils import misc
 from pyhanko.pdf_utils.config_utils import ConfigurationError
 from pyhanko.pdf_utils.crypt import (
     SimpleEnvelopeKeyDecrypter,
-    PubKeySecurityHandler, AuthResult,
+    PubKeySecurityHandler, AuthStatus,
 )
 
-from pyhanko.sign import signers
+from pyhanko.sign import signers, pkcs11
 from pyhanko.sign.general import SigningError
 from pyhanko.sign.timestamps import HTTPTimeStamper
 from pyhanko.sign import validation, beid, fields
@@ -504,9 +504,17 @@ def _open_for_signing(infile_path, signer_cert=None, signer_key=None):
     return writer
 
 
+def get_text_params(ctx):
+    text_params = None
+    stamp_url = ctx.obj[Ctx.STAMP_URL]
+    if stamp_url is not None:
+        text_params = {'url': stamp_url}
+    return text_params
+
+
 def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
                          timestamp_url, signature_meta, existing_fields_only,
-                         style, stamp_url, new_field_spec):
+                         style, text_params, new_field_spec):
     with pyhanko_exception_manager():
         if timestamp_url is not None:
             timestamper = HTTPTimeStamper(timestamp_url)
@@ -517,24 +525,30 @@ def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
             signer_key=signer.signing_key
         )
 
-        text_params = None
-        if stamp_url is not None:
-            text_params = {'url': stamp_url}
-
-        result = signers.PdfSigner(
-            signature_meta, signer=signer, timestamper=timestamper,
-            stamp_style=style, new_field_spec=new_field_spec
-        ).sign_pdf(
-            writer, existing_fields_only=existing_fields_only,
-            appearance_text_params=text_params
+        generic_sign(
+            writer=writer, outfile=outfile,
+            signature_meta=signature_meta, signer=signer,
+            timestamper=timestamper, style=style, new_field_spec=new_field_spec,
+            existing_fields_only=existing_fields_only, text_params=text_params
         )
 
-        buf = result.getbuffer()
-        outfile.write(buf)
-        buf.release()
 
-        writer.prev.stream.close()
-        outfile.close()
+def generic_sign(*, writer, outfile, signature_meta, signer, timestamper,
+                 style, new_field_spec, existing_fields_only, text_params):
+    result = signers.PdfSigner(
+        signature_meta, signer=signer, timestamper=timestamper,
+        stamp_style=style, new_field_spec=new_field_spec
+    ).sign_pdf(
+        writer, existing_fields_only=existing_fields_only,
+        appearance_text_params=text_params
+    )
+
+    buf = result.getbuffer()
+    outfile.write(buf)
+    buf.release()
+
+    writer.prev.stream.close()
+    outfile.close()
 
 
 @addsig.command(name='pemder', help='read key material from PEM/DER files')
@@ -573,7 +587,7 @@ def addsig_pemder(ctx, infile, outfile, key, cert, chain, passfile):
         signer, infile, outfile, timestamp_url=timestamp_url,
         signature_meta=signature_meta,
         existing_fields_only=existing_fields_only,
-        style=ctx.obj[Ctx.STAMP_STYLE], stamp_url=ctx.obj[Ctx.STAMP_URL],
+        style=ctx.obj[Ctx.STAMP_STYLE], text_params=get_text_params(ctx),
         new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC]
     )
 
@@ -614,11 +628,63 @@ def addsig_pkcs12(ctx, infile, outfile, pfx, chain, passfile):
         signer, infile, outfile, timestamp_url=timestamp_url,
         signature_meta=signature_meta,
         existing_fields_only=existing_fields_only,
-        style=ctx.obj[Ctx.STAMP_STYLE], stamp_url=ctx.obj[Ctx.STAMP_URL],
+        style=ctx.obj[Ctx.STAMP_STYLE], text_params=get_text_params(ctx),
         new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC]
     )
 
 
+# TODO add options to specify extra certs to include
+
+@addsig.command(name='pkcs11', help='use generic PKCS#11 device to sign')
+@click.argument('infile', type=click.File('rb'))
+@click.argument('outfile', type=click.File('wb'))
+@click.option('--lib', help='path to PKCS#11 module',
+              type=readable_file, required=True)
+@click.option('--token-label', help='PKCS#11 token label', type=str,
+              required=True)
+@click.option('--cert-label', help='certificate label', type=str, required=True)
+@click.option('--key-label', help='key label', type=str, required=False)
+@click.option('--slot-no', help='specify PKCS#11 slot to use',
+              required=False, type=int, default=None)
+@click.option('--skip-user-pin', type=bool, show_default=True,
+              default=False, required=False, is_flag=True,
+              help='do not prompt for PIN (e.g. if the token has a PIN pad)')
+@click.pass_context
+def addsig_pkcs11(ctx, infile, outfile, lib, token_label,
+                  cert_label, key_label, slot_no, skip_user_pin):
+    signature_meta = ctx.obj[Ctx.SIG_META]
+    existing_fields_only = ctx.obj[Ctx.EXISTING_ONLY]
+    timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
+
+    if skip_user_pin:
+        user_pin = None
+    else:
+        user_pin = getpass.getpass(prompt='PKCS#11 user PIN: ')
+
+    session = pkcs11.open_pkcs11_session(
+        lib_location=lib, slot_no=slot_no, token_label=token_label,
+        user_pin=user_pin
+    )
+    if timestamp_url is not None:
+        timestamper = HTTPTimeStamper(timestamp_url)
+    else:
+        timestamper = None
+
+    signer = pkcs11.PKCS11Signer(
+        session, cert_label=cert_label, key_label=key_label,
+    )
+
+    with pyhanko_exception_manager():
+        generic_sign(
+            writer=IncrementalPdfFileWriter(infile), outfile=outfile,
+            signature_meta=signature_meta, signer=signer,
+            timestamper=timestamper, style=ctx.obj[Ctx.STAMP_STYLE],
+            new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
+            existing_fields_only=existing_fields_only,
+            text_params=get_text_params(ctx)
+        )
+
+
 @addsig.command(name='beid', help='use Belgian eID to sign')
 @click.argument('infile', type=click.File('rb'))
 @click.argument('outfile', type=click.File('wb'))
@@ -635,36 +701,22 @@ def addsig_beid(ctx, infile, outfile, lib, use_auth_cert, slot_no):
     existing_fields_only = ctx.obj[Ctx.EXISTING_ONLY]
     timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
     session = beid.open_beid_session(lib, slot_no=slot_no)
-    label = 'Authentication' if use_auth_cert else 'Signature'
     if timestamp_url is not None:
         timestamper = HTTPTimeStamper(timestamp_url)
     else:
         timestamper = None
-    signer = beid.BEIDSigner(
-        session, label
-    )
 
-    stamp_url = ctx.obj[Ctx.STAMP_URL]
-    text_params = None
-    if stamp_url is not None:
-        text_params = {'url': stamp_url}
+    signer = beid.BEIDSigner(session, use_auth_cert=use_auth_cert)
 
     with pyhanko_exception_manager():
-        writer = IncrementalPdfFileWriter(infile)
-        result = signers.PdfSigner(
-            signature_meta, signer=signer, timestamper=timestamper,
-            stamp_style=ctx.obj[Ctx.STAMP_STYLE],
-            new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC]
-        ).sign_pdf(
-            writer, existing_fields_only=existing_fields_only,
-            appearance_text_params=text_params
+        generic_sign(
+            writer=IncrementalPdfFileWriter(infile), outfile=outfile,
+            signature_meta=signature_meta, signer=signer,
+            timestamper=timestamper, style=ctx.obj[Ctx.STAMP_STYLE],
+            new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
+            existing_fields_only=existing_fields_only,
+            text_params=get_text_params(ctx)
         )
-        buf = result.getbuffer()
-        outfile.write(buf)
-        buf.release()
-
-        infile.close()
-        outfile.close()
 
 
 def _index_page(page):
@@ -842,13 +894,13 @@ def decrypt_with_password(infile, outfile, password, force):
             if not password:
                 password = getpass.getpass(prompt='File password: ')
             auth_result = r.decrypt(password)
-            if auth_result == AuthResult.USER and not force:
+            if auth_result.status == AuthStatus.USER and not force:
                 raise click.ClickException(
                     "Password specified was the user password, not "
                     "the owner password. Pass --force to decrypt the "
                     "file anyway."
                 )
-            elif auth_result == AuthResult.FAILED:
+            elif auth_result.status == AuthStatus.FAILED:
                 raise click.ClickException("Password didn't match.")
             w = copy_into_new_writer(r)
             with open(outfile, 'wb') as outf:
@@ -888,7 +940,7 @@ def _decrypt_pubkey(sedk: SimpleEnvelopeKeyDecrypter, infile, outfile, force):
                     "File was not encrypted with a public-key security handler."
                 )
             auth_result = r.decrypt_pubkey(sedk)
-            if auth_result == AuthResult.USER:
+            if auth_result.status == AuthStatus.USER:
                 # TODO read 2nd bit of perms in CMS enveloped data
                 #  is the one indicating that change of encryption is OK
                 if not force:
@@ -896,7 +948,7 @@ def _decrypt_pubkey(sedk: SimpleEnvelopeKeyDecrypter, infile, outfile, force):
                         "Change of encryption is typically not allowed with "
                         "user access. Pass --force to decrypt the file anyway."
                     )
-            elif auth_result == AuthResult.FAILED:
+            elif auth_result.status == AuthStatus.FAILED:
                 raise click.ClickException("Failed to decrypt the file.")
             w = copy_into_new_writer(r)
             with open(outfile, 'wb') as outf:
