@@ -79,9 +79,6 @@ class XRefCache:
         self.xref_locations = []
         self.in_obj_stream = {}
         self.standard_xrefs = {}
-        # keep track of the xref section that last changed an entry
-        #  (needed for some validation workflows)
-        self.last_change = {}
         # making this a dict doesn't make much sense
         self.history = defaultdict(list)
         self._current_section_ids = set()
@@ -92,6 +89,11 @@ class XRefCache:
         self._previous_expected_free = {}
         self.xref_container_info = []
 
+        # Objects that were declared as 'xxxxxxx 00000 f' in the
+        # initial revision. This sometimes happens when PDF writers clean up
+        # dead objects in a file, but want to preserve existing object IDs
+        self._initially_dead_objects = set()
+
         self._obj_streams_by_revision = defaultdict(set)
 
     def _next_section(self):
@@ -111,9 +113,40 @@ class XRefCache:
     def free_ref(self, idnum, next_generation):
         if not idnum:
             return
+
+        # When rewriting files & removing dead objects, Acrobat will
+        # enter the deleted reference into the Xref table/stream with
+        # a 'next generation' ID of 0. It doesn't contradict the spec directly,
+        # but I assumed that this was the way to indicate that generation 0xffff
+        # had just been freed. Apparently not, because I've seen Acrobat
+        # put that same freed reference in later revisions with a next_gen
+        # number of 1. Bizarre.
+        #
+        # Anyhow, given the ubiquity of Adobe (Acrobat|Reader), it's probably
+        # prudent to special-case this one.
+        # In doing so, we're probably not dealing correctly with the case
+        # where the 0xffff'th generation of an object is freed, but I'm happy
+        # to assume that that will never happen in a legitimate file.
+        if not next_generation:
+            self._initially_dead_objects.add(idnum)
+            # remove any subsequent freeings of the 0th generation, since it
+            # never existed in the first place
+            zeroth_gen_ref = generic.Reference(idnum, 0)
+            later_revs = zip(self._freed_by_section, self._refs_by_section)
+            for freed, defd in later_revs:
+                try:
+                    freed.remove(zeroth_gen_ref)
+                except KeyError:
+                    continue  # not freed in this revision, move on
+
+                # also delete the record of the ref being modified in that
+                # generation (freed is a subset of defd by construction)
+                defd.remove(zeroth_gen_ref)
+            return
+
         # treat this as setting idnum, next_generation-1 to null
-        prev_generation = (next_generation - 1) if next_generation else 0xffff
-        self.standard_xrefs[(prev_generation, idnum)] = generic.NullObject()
+        prev_generation = next_generation - 1
+        self.standard_xrefs[(prev_generation, idnum)] = None
         null_ref = generic.Reference(idnum, prev_generation)
         self._current_section_freed.add(null_ref)
         self._current_section_ids.add(null_ref)
@@ -133,9 +166,6 @@ class XRefCache:
         except StopIteration:
             self._generations[idnum].add(prev_generation)
 
-        if idnum not in self.last_change:
-            # this revision is the last change
-            self.last_change[idnum] = self.xref_sections
         try:
             # remove from expected free dict
             expected_generation = self._previous_expected_free.pop(idnum)
@@ -150,6 +180,12 @@ class XRefCache:
             pass
 
     def put_ref(self, idnum, generation, start):
+        if idnum in self._initially_dead_objects:
+            # see comments in free_ref for justification
+            raise PdfReadError(
+                f"Spurious history for object {idnum}; is treated as dead "
+                f"reference later in file."
+            )
         if idnum in self._previous_expected_free:
             raise PdfReadError(
                 f"Generation {generation} of object {idnum} was "
@@ -164,7 +200,6 @@ class XRefCache:
             self._previous_expected_free[idnum] = generation
         if not self.used_later(idnum, generation):
             self.standard_xrefs[(generation, idnum)] = start
-            self.last_change[idnum] = self.xref_sections
             self._generations[idnum] = {generation}
         else:
             self._generations[idnum].add(generation)
@@ -180,7 +215,6 @@ class XRefCache:
         marker = (obj_stream_num, obj_stream_ix)
         if not self.used_later(idnum, 0):
             self.in_obj_stream[idnum] = marker
-            self.last_change[idnum] = self.xref_sections
             self._generations[idnum] = {0}
 
         self.history[(0, idnum)].append((self.xref_sections, marker))
@@ -190,8 +224,10 @@ class XRefCache:
     def total_revisions(self):
         return self.xref_sections
 
-    def get_last_change(self, idnum):
-        return self.xref_sections - 1 - self.last_change[idnum]
+    def get_last_change(self, ref: generic.Reference):
+        ref_hist = self.history[(ref.generation, ref.idnum)]
+        section, _ = ref_hist[0]
+        return self.xref_sections - 1 - section
 
     def object_streams_used_in(self, revision):
         return self._obj_streams_by_revision[self.xref_sections - 1 - revision]
@@ -733,9 +769,7 @@ class PdfFileReader(PdfHandler):
 
     def _read_object(self, ref, marker, never_decrypt=False):
         if marker is None:
-            raise PdfReadError(
-                f"Reference {ref} has been freed."
-            )
+            return generic.NullObject()
         elif isinstance(marker, tuple):
             # object in object stream
             (obj_stream_num, obj_stream_ix) = marker
@@ -1011,8 +1045,8 @@ class PdfFileReader(PdfHandler):
 
         result = sorted(
             (
-                EmbeddedPdfSignature(self, sig_field)
-                for _, sig_obj, sig_field in sig_fields
+                EmbeddedPdfSignature(self, sig_field, fq_name)
+                for fq_name, sig_obj, sig_field in sig_fields
             ), key=lambda emb: emb.signed_revision
         )
         self._embedded_signatures = result
@@ -1054,6 +1088,51 @@ class RawPdfPath:
             and (self is other or self._tag() == other._tag())
         )
 
+    def access_on(self, from_obj, dereference_last=True) -> generic.PdfObject:
+        current_obj = from_obj
+        for ix, entry in enumerate(self.path):
+            # we put this here to make dereference_last work
+            if isinstance(current_obj, generic.IndirectObject):
+                current_obj = current_obj.get_object()
+            if isinstance(entry, str):
+                if isinstance(current_obj, generic.DictionaryObject):
+                    try:
+                        current_obj = current_obj.raw_get(entry)
+                        continue
+                    except KeyError:
+                        raise misc.PdfReadError(
+                            f"Encountered missing dictionary "
+                            f"entry {entry} at position {ix} in path {self}"
+                            f"from {from_obj}."
+                        )
+            elif isinstance(entry, int):
+                if isinstance(current_obj, generic.ArrayObject):
+                    if not (0 <= entry <= len(current_obj)):
+                        raise misc.PdfReadError(
+                            f"Encountered out-of-range array index "
+                            f"{entry} at position {ix} in path {self}"
+                            f"from {from_obj}."
+                        )
+                    current_obj = current_obj.raw_get(entry)
+                    continue
+            # if we get here, there's a typing issue
+            raise misc.PdfReadError(
+                f"Type error in path {self} at position {ix}."
+            )
+        if isinstance(current_obj, generic.IndirectObject) and dereference_last:
+            return current_obj.get_object()
+        else:
+            return current_obj
+
+    def access_reference_on(self, from_obj) -> generic.Reference:
+        ind_obj = self.access_on(from_obj, dereference_last=False)
+        if not isinstance(ind_obj, generic.IndirectObject):
+            raise misc.IndirectObjectExpected(
+                f"Final entity on path {self} starting from {from_obj} is not "
+                f"an indirect object."
+            )
+        return ind_obj.reference
+
     @staticmethod
     def _fmt_node(node):
         if isinstance(node, int):
@@ -1122,7 +1201,7 @@ class HistoricalResolver(PdfHandler):
             # we can grab it from the "normal" shared cache.
             reader = self.reader
             revision = self.revision
-            if reader.xrefs.get_last_change(ref.idnum) <= revision:
+            if reader.xrefs.get_last_change(ref) <= revision:
                 obj = reader.get_object(ref)
             else:
                 obj = reader.get_object(ref, revision)
@@ -1257,13 +1336,15 @@ class HistoricalResolver(PdfHandler):
         # We flatten everything when we're done
         def _compute_paths_to_refs(obj, cur_path: misc.ConsList,
                                    seen_in_path: misc.ConsList, *,
-                                   is_page_tree, page_tree_objs):
+                                   is_page_tree, page_tree_objs,
+                                   is_struct_tree, struct_tree_objs):
 
             # optimisation: page tree gets special treatment
             # to prevent unnecessary paths from being generated when the
             # tree is entered from the outside (e.g. by following /P on a form
             # field/widget annotation)
-            # TODO what about the structure tree?
+            # The structure tree is similarly special-cased.
+
             # Can we deal with this more systematically? The problem is that
             # there's no a priori way to state which path from an object
             # to the trailer is the "canonical" one. For page objects
@@ -1279,11 +1360,13 @@ class HistoricalResolver(PdfHandler):
                 obj = self(ref)
                 if not is_page_tree and ref in page_tree_objs:
                     return
+                if not is_struct_tree and ref in struct_tree_objs:
+                    return
             if isinstance(obj, generic.DictionaryObject):
                 for k, v in obj.items():
                     # another hack to eliminate some spurious extra paths
                     # that don't convey any useful information
-                    if k == '/Parent':
+                    if k == '/Parent' or (is_struct_tree and k == '/P'):
                         continue
 
                     _compute_paths_to_refs(
@@ -1292,14 +1375,28 @@ class HistoricalResolver(PdfHandler):
                             cur_path.head == '/Root' and k == '/Pages'
                             and cur_path.tail == misc.ConsList.empty()
                         ),
-                        page_tree_objs=page_tree_objs
+                        page_tree_objs=page_tree_objs,
+                        # for the struct tree: we definitely want to
+                        # consider the /ParentTree as an "external" feature
+                        # here, since it contains lots of references to
+                        # structure elements that only exist for indexing
+                        # purposes so they don't add any extra information
+                        # (in the sense that recursing into them accomplishes
+                        # nothing)
+                        is_struct_tree=is_struct_tree or (
+                            cur_path.head == '/StructTreeRoot' and k == '/K'
+                            and cur_path.tail == misc.ConsList.sing('/Root')
+                        ),
+                        struct_tree_objs=struct_tree_objs
                     )
             elif isinstance(obj, generic.ArrayObject):
                 for ix, v in enumerate(obj):
                     _compute_paths_to_refs(
                         v, cur_path.cons(ix), seen_in_path,
                         is_page_tree=is_page_tree,
-                        page_tree_objs=page_tree_objs
+                        page_tree_objs=page_tree_objs,
+                        is_struct_tree=is_struct_tree,
+                        struct_tree_objs=struct_tree_objs
                     )
 
         def _collect_page_tree_refs(pages_obj):
@@ -1311,14 +1408,56 @@ class HistoricalResolver(PdfHandler):
                 if kid.get('/Type', None) == '/Pages':
                     yield from _collect_page_tree_refs(kid)
 
+        def _collect_struct_tree_refs(struct_elem):
+            try:
+                children = struct_elem['/K']
+            except KeyError:
+                return
+
+            # if there's only one child, /K need not be an array
+            if not isinstance(children, generic.ArrayObject):
+                children = children,
+
+            for child in children:
+                child_ref = None
+                if isinstance(child, generic.IndirectObject):
+                    child_ref = child.reference
+                    child = child.get_object()
+
+                # The /K entry can also refer to content items.
+                # We don't care about those.
+                if not isinstance(child, generic.DictionaryObject):
+                    continue
+                try:
+                    if child['/Type'] != '/StructElem':
+                        continue
+                except KeyError:
+                    # If the child doesn't have a /Type entry, /StructElem
+                    # is the default (says so in the spec)
+                    pass
+
+                if child_ref is not None:
+                    yield child_ref
+
+                yield from _collect_struct_tree_refs(child)
+
         pages_ref = self.root.raw_get('/Pages')
         page_tree_nodes = set()
         page_tree_nodes.update(
             _collect_page_tree_refs(pages_obj=pages_ref.get_object())
         )
+        struct_tree_nodes = set()
+        try:
+            struct_tree_root_ref = self.root.raw_get('/StructTreeRoot')
+            struct_tree_nodes.update(
+                _collect_struct_tree_refs(struct_tree_root_ref.get_object())
+            )
+        except KeyError:
+            pass
         _compute_paths_to_refs(
             self.trailer_view, misc.ConsList.empty(), misc.ConsList.empty(),
-            is_page_tree=False, page_tree_objs=page_tree_nodes
+            is_page_tree=False, page_tree_objs=page_tree_nodes,
+            is_struct_tree=False, struct_tree_objs=struct_tree_nodes
         )
 
         self._indirect_object_access_cache = {
