@@ -75,7 +75,7 @@ from pyhanko.sign.fields import FieldMDPSpec, MDPPerm
 
 __all__ = [
     'ModificationLevel', 'SuspiciousModification',
-    'QualifiedWhitelistRule', 'WhitelistRule', 'qualify',
+    'QualifiedWhitelistRule', 'WhitelistRule', 'qualify', 'ReferenceUpdate',
     'DocInfoRule', 'DSSCompareRule', 'MetadataUpdateRule',
     'CatalogModificationRule', 'ObjectStreamRule', 'XrefStreamRule',
     'FormUpdatingRule', 'FormUpdate',
@@ -531,14 +531,54 @@ class FieldComparisonSpec:
         assert isinstance(field, generic.DictionaryObject)
         return field
 
-    def old_annotation_paths(self):
+    def expected_paths(self):
+        # these are the paths where we expect the form field to be referred to
+        paths = self._old_annotation_paths()
+        struct_path = self._find_in_structure_tree()
+        if struct_path is not None:
+            paths.add(struct_path)
+        paths.add(self.old_canonical_path)
+        return paths
+
+    def _find_in_structure_tree(self):
+        # collect paths (0 or 1) through which this field appears
+        #  in the file's structure tree.
+
+        # TODO check whether the structure element is a form control
+        #  (or something role-mapped to it)
+        # TODO if multiple paths exist, we should only whitelist the one
+        #  that corresponds to the StructParent entry, not just the first one
+
+        # Note: the path simplifier suppresses the extra cross-references
+        # from parent pointers in the tree and from the /ParentTree index.
+
+        old_field_ref = self.old_field_ref
+        old = self.old_field_ref.get_pdf_handler()
+        assert isinstance(old, HistoricalResolver)
+
+        if '/StructTreeRoot' not in old.root:
+            return
+
+        # check if the path ends in Form.K.Obj and
+        # starts with Root.StructTreeRoot.K
+        for pdf_path in old._get_usages_of_ref(old_field_ref):
+            # Root.StructTreeRoot.K is three, and K.Obj at the end is
+            # another 2
+            if '/StructTreeRoot' in pdf_path and len(pdf_path) >= 5:
+                root, struct_tree_root, k1 = pdf_path.path[:3]
+                k2, obj = pdf_path.path[-2:]
+                if k1 == k2 == '/K' and obj == '/Obj' and root == '/Root' \
+                        and struct_tree_root == '/StructTreeRoot':
+                    return pdf_path
+
+    def _old_annotation_paths(self):
         # collect path(s) through which this field is used as an annotation
         # the clean way to accomplish this would be to follow /P
         # and go from there, but /P is optional, so we have to get a little
         # creative.
         old_field_ref = self.old_field_ref
         if old_field_ref is None:
-            return set()
+            return set()  # pragma: nocover
 
         old = self.old_field_ref.get_pdf_handler()
         assert isinstance(old, HistoricalResolver)
@@ -667,10 +707,6 @@ class SigFieldCreationRule(FieldMDPRule):
             for fq_name, spec in context.field_specs.items():
                 if spec.field_type != '/Sig' or spec.old_field_ref:
                     continue
-                if '.' in fq_name:
-                    raise NotImplementedError(
-                        "Can't deal with signature fields that aren't top level"
-                    )
                 yield fq_name, spec.new_field_ref
 
         all_new_refs = dict(_collect())
@@ -874,14 +910,10 @@ class SigFieldModificationRule(BaseFieldModificationRule):
             # permissible even when the field is locked.
             valid_when_locked = self.compare_fields(spec)
 
-            # these are the paths where we expect to encounter this signature
-            # field
-            paths = spec.old_annotation_paths()
-            paths.add(spec.old_canonical_path)
             field_ref_update = FormUpdate(
                 updated_ref=spec.new_field_ref, field_name=fq_name,
                 valid_when_locked=valid_when_locked,
-                paths_checked=paths
+                paths_checked=spec.expected_paths()
             )
 
             if not previously_signed and now_signed:
@@ -942,12 +974,15 @@ class SigFieldModificationRule(BaseFieldModificationRule):
         # enforce as a rigid rule here)
         if sig_obj.raw_get('/Type') == '/DocTimeStamp' and not area:
             sig_whitelist = ModificationLevel.LTA_UPDATES
+            valid_when_locked = True
         else:
             sig_whitelist = ModificationLevel.FORM_FILLING
+            valid_when_locked = False
 
         # first, whitelist the actual signature object
         yield sig_whitelist, FormUpdate(
-            updated_ref=current_value_ref, field_name=fq_name
+            updated_ref=current_value_ref, field_name=fq_name,
+            valid_when_locked=valid_when_locked
         )
 
         # since apparently Acrobat didn't get the memo about not having
@@ -994,14 +1029,12 @@ class GenericFieldModificationRule(BaseFieldModificationRule):
 
         valid_when_locked = self.compare_fields(spec)
 
-        # these are the paths where we expect the form field to be referred to
-        paths = spec.old_annotation_paths()
-        paths.add(spec.old_canonical_path)
         yield (
             ModificationLevel.FORM_FILLING,
             FormUpdate(
                 updated_ref=spec.new_field_ref, field_name=fq_name,
-                valid_when_locked=valid_when_locked, paths_checked=paths
+                valid_when_locked=valid_when_locked,
+                paths_checked=spec.expected_paths()
             )
         )
         old_field = spec.old_field
@@ -1360,17 +1393,18 @@ def _list_fields(old_fields: generic.PdfObject, new_fields: generic.PdfObject,
             try:
                 field_type = field.raw_get('/FT')
             except KeyError:
-                if inherited_ft is not None:
-                    field_type = inherited_ft
-                else:
-                    raise exc(
-                        f"Field type of {fq_name} could not be determined"
-                    )
+                field_type = inherited_ft
 
             try:
                 kids = field["/Kids"]
             except KeyError:
                 kids = generic.ArrayObject()
+
+            if not kids and field_type is None:
+                raise exc(
+                    f"Field type of terminal field {fq_name} could not be "
+                    f"determined"
+                )
             yield fq_name, (field_type, field_ref.reference, kids, ix)
 
     old_fields_by_name = dict(_make_list(old_fields, misc.PdfReadError))
