@@ -15,14 +15,13 @@ from oscrypto import keys as oskeys
 
 from pyhanko.pdf_utils import generic
 from pyhanko.pdf_utils.generic import pdf_name, pdf_string
-from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
 from pyhanko.pdf_utils.misc import OrderedEnum, PdfWriteError, get_and_apply
 from pyhanko.pdf_utils.rw_common import PdfHandler
+from pyhanko.pdf_utils.writer import BasePdfFileWriter
 from pyhanko.sign.general import (
     UnacceptableSignerError, SigningError,
     KeyUsageConstraints,
 )
-from pyhanko.stamp import AnnotAppearances
 
 __all__ = [
     'SigFieldSpec', 'SigSeedValFlags', 'SigCertConstraints',
@@ -1187,8 +1186,50 @@ class SigFieldSpec:
         return result
 
 
+def _insert_or_get_field_at(writer: BasePdfFileWriter, fields, path,
+                            parent_ref=None, modified=False, field_obj=None):
+
+    current_partial, tail = path[0], path[1:]
+
+    for field_ref in fields:
+        assert isinstance(field_ref, generic.IndirectObject)
+        field = field_ref.get_object()
+        if field.get('/T', None) == current_partial:
+            break
+    else:
+        # have to insert a new element into the fields array
+        if field_obj is not None and not tail:
+            field = field_obj
+        else:
+            # create a generic field
+            field = generic.DictionaryObject()
+        field['/T'] = pdf_string(current_partial)
+        if parent_ref is not None:
+            field['/Parent'] = parent_ref
+        field_ref = writer.add_object(field)
+        fields.append(field_ref)
+        writer.update_container(fields)
+        modified = True
+
+    if not tail:
+        return modified, field_ref
+    # check for /Kids, and create it if necessary
+    try:
+        kids = field['/Kids']
+    except KeyError:
+        kids = field['/Kids'] = generic.ArrayObject()
+        writer.update_container(field)
+        modified = True
+
+    # recurse in to /Kids array
+    return _insert_or_get_field_at(
+        writer, kids, tail, parent_ref=field_ref,
+        modified=modified, field_obj=field_obj
+    )
+
+
 def _prepare_sig_field(sig_field_name, root,
-                       update_writer: IncrementalPdfFileWriter,
+                       update_writer: BasePdfFileWriter,
                        existing_fields_only=False, lock_sig_flags=True,
                        **kwargs):
     """
@@ -1239,22 +1280,17 @@ def _prepare_sig_field(sig_field_name, root,
     if sig_field_ref is not None:
         return False, sig_field_ref
 
-    if '.' in sig_field_name:
-        raise NotImplementedError(
-            "Creating fields deep in the form hierarchy is not supported"
-            "right now."
-        )
-
     # no signature field exists, so create one
     # default: grab a reference to the first page
     page_ref = update_writer.find_page_for_modification(0)[0]
     sig_form_kwargs = {'include_on_page': page_ref}
     sig_form_kwargs.update(**kwargs)
-    sig_field = SignatureFormField(
-        sig_field_name, writer=update_writer, **sig_form_kwargs
+    sig_field = SignatureFormField(sig_field_name, **sig_form_kwargs)
+    created, sig_field_ref = _insert_or_get_field_at(
+        update_writer, fields, path=sig_field_name.split('.'),
+        field_obj=sig_field
     )
-    sig_field_ref = sig_field.reference
-    fields.append(sig_field_ref)
+    update_writer.register_annotation(page_ref, sig_field_ref)
 
     # make sure /SigFlags is present. If not, create it
     sig_flags = 3 if lock_sig_flags else 1
@@ -1354,13 +1390,13 @@ def enumerate_sig_fields_in(field_list, filled_status=None, with_name=None,
             try:
                 yield from enumerate_sig_fields_in(
                     field['/Kids'], parent_name=fq_name, parents=current_path,
-                    with_name=with_name
+                    with_name=with_name, filled_status=filled_status
                 )
             except KeyError:
                 continue
 
 
-def append_signature_field(pdf_out: IncrementalPdfFileWriter,
+def append_signature_field(pdf_out: BasePdfFileWriter,
                            sig_field_spec: SigFieldSpec):
     """
     Append signature fields to a PDF file.
@@ -1401,43 +1437,26 @@ def append_signature_field(pdf_out: IncrementalPdfFileWriter,
 
 
 class SignatureFormField(generic.DictionaryObject):
-    def __init__(self, field_name, include_on_page, *, writer,
-                 sig_object_ref=None, box=None,
-                 appearances: Optional[AnnotAppearances] = None):
+    def __init__(self, field_name, *, box=None, include_on_page=None,
+                 is_annotation=True,
+                 # this sets the "print" bit
+                 annot_flags=0b100):
 
         if box is not None:
-            visible = True
             rect = list(map(generic.FloatObject, box))
-            if appearances is not None:
-                ap = appearances.as_pdf_object()
-            else:
-                ap = None
         else:
             rect = [generic.FloatObject(0)] * 4
-            ap = None
-            visible = False
 
-        # this sets the "Print" bit, and activates "Locked" if the
-        # signature field is ready to be filled
-        flags = 0b100 if sig_object_ref is None else 0b10000100
         super().__init__({
             # Signature field properties
             pdf_name('/FT'): pdf_name('/Sig'),
             pdf_name('/T'): pdf_string(field_name),
-            # Annotation properties: bare minimum
-            pdf_name('/Type'): pdf_name('/Annot'),
-            pdf_name('/Subtype'): pdf_name('/Widget'),
-            pdf_name('/F'): generic.NumberObject(flags),
-            pdf_name('/P'): include_on_page,
-            pdf_name('/Rect'): generic.ArrayObject(rect)
         })
-        if sig_object_ref is not None:
-            self[pdf_name('/V')] = sig_object_ref
-        if ap is not None:
-            self[pdf_name('/AP')] = ap
-
-        # register ourselves
-        self.reference = self_reference = writer.add_object(self)
-        # if we're building an invisible form field, this is all there is to it
-        if visible:
-            writer.register_annotation(include_on_page, self_reference)
+        if is_annotation:
+            # Annotation properties: bare minimum
+            self['/Type'] = pdf_name('/Annot')
+            self['/Subtype'] = pdf_name('/Widget')
+            self['/F'] = generic.NumberObject(annot_flags)
+            self['/Rect'] = generic.ArrayObject(rect)
+            if include_on_page is not None:
+                self['/P'] = include_on_page
