@@ -20,6 +20,7 @@ from pyhanko.pdf_utils.generic import pdf_name, pdf_date, pdf_string
 from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
 from pyhanko.pdf_utils.layout import BoxConstraints
 from pyhanko.pdf_utils.reader import PdfFileReader
+from pyhanko.pdf_utils.writer import BasePdfFileWriter
 from pyhanko.sign import general
 from pyhanko.sign.fields import (
     enumerate_sig_fields, _prepare_sig_field,
@@ -40,7 +41,8 @@ __all__ = [
     'PdfSignatureMetadata',
     'Signer', 'SimpleSigner', 'PdfTimeStamper', 'PdfSigner',
     'PdfCMSEmbedder', 'SigObjSetup', 'SigAppearanceSetup', 'SigMDPSetup',
-    'PdfSignedData', 'SignatureObject', 'DocumentTimestamp',
+    'PdfByteRangeDigest', 'PdfSignedData',
+    'SignatureObject', 'DocumentTimestamp',
     'SigIOSetup', 'sign_pdf', 'load_certs_from_pemder',
     'DEFAULT_MD', 'DEFAULT_SIGNING_STAMP_STYLE', 'DEFAULT_SIG_SUBFILTER'
 ]
@@ -60,11 +62,11 @@ class SigByteRangeObject(generic.PdfObject):
 
     def fill_offsets(self, stream, sig_start, sig_end, eof):
         if self._filled:
-            raise ValueError('Offsets already filled')
+            raise ValueError('Offsets already filled')  # pragma: nocover
         if self._range_object_offset is None:
             raise ValueError(
                 'Could not determine where to write /ByteRange value'
-            )
+            )  # pragma: nocover
 
         old_seek = stream.tell()
         self.first_region_len = sig_start
@@ -99,13 +101,9 @@ class DERPlaceholder(generic.PdfObject):
     @property
     def offsets(self):
         if self._offsets is None:
-            raise ValueError('No offsets available')
+            raise ValueError('No offsets available')  # pragma: nocover
         return self._offsets
 
-    @property
-    def original_bytes(self):
-        return self.value
-
     # always ignore encryption key, since this is a placeholder
     def write_to_stream(self, stream, handler=None, container_ref=None):
         start = stream.tell()
@@ -120,52 +118,21 @@ class DERPlaceholder(generic.PdfObject):
 DEFAULT_SIG_SUBFILTER = SigSeedSubFilter.ADOBE_PKCS7_DETACHED
 
 
-class PdfSignedData(generic.DictionaryObject):
-    """
-    Generic class to model signature dictionaries in a PDF file.
-    See also :class:`.SignatureObject` and :class:`.DocumentTimestamp`.
+class PdfByteRangeDigest(generic.DictionaryObject):
 
-    :param obj_type:
-        The type of signature object.
-    :param subfilter:
-        See :class:`.SigSeedSubFilter`.
-    :param timestamp:
-        The timestamp to embed into the ``/M`` entry.
-    :param bytes_reserved:
-        The number of bytes to reserve for the signature.
-        Defaults to 16 KiB.
-
-        .. warning::
-            Since the CMS object is written to the output file as a hexadecimal
-            string, you should request **twice** the (estimated) number of bytes
-            in the DER-encoded version of the CMS object.
-    """
-
-    def __init__(self, obj_type,
-                 subfilter: SigSeedSubFilter = DEFAULT_SIG_SUBFILTER,
-                 timestamp: datetime = None, bytes_reserved=None):
+    def __init__(self, data_key=pdf_name('/Contents'), *, bytes_reserved=None):
+        super().__init__()
         if bytes_reserved is not None and bytes_reserved % 2 == 1:
             raise ValueError('bytes_reserved must be even')
 
-        super().__init__(
-            {
-                pdf_name('/Type'): obj_type,
-                pdf_name('/Filter'): pdf_name('/Adobe.PPKLite'),
-                pdf_name('/SubFilter'): subfilter.value,
-            }
-        )
-
-        if timestamp is not None:
-            self[pdf_name('/M')] = pdf_date(timestamp)
-
-        # initialise placeholders for /Contents and /ByteRange
-        sig_contents = DERPlaceholder(bytes_reserved=bytes_reserved)
-        self[pdf_name('/Contents')] = self.signature_contents = sig_contents
+        self.data_key = data_key
+        contents = DERPlaceholder(bytes_reserved=bytes_reserved)
+        self[data_key] = self.contents = contents
         byte_range = SigByteRangeObject()
         self[pdf_name('/ByteRange')] = self.byte_range = byte_range
 
-    def write_signature(self, writer: IncrementalPdfFileWriter, md_algorithm,
-                        in_place=False, output=None, chunk_size=4096):
+    def fill(self, writer: BasePdfFileWriter, md_algorithm,
+             in_place=False, output=None, chunk_size=4096):
         """
         Generator coroutine that handles the document hash computation and
         the actual filling of the placeholder data.
@@ -174,19 +141,39 @@ class PdfSignedData(generic.DictionaryObject):
         wherever possible. If you *really* need fine-grained control,
         use :class:`.PdfCMSEmbedder` instead.
         """
+
         if in_place:
+            if not isinstance(writer, IncrementalPdfFileWriter):
+                raise TypeError(
+                    "in_place is only meaningful for incremental writers."
+                )  # pragma: nocover
             output = writer.prev.stream
             writer.write_in_place()
         else:
             # Render the PDF to a byte buffer with placeholder values
-            # for the signature data
-            output = BytesIO() if output is None else output
+            # for the signature data, or straight to the provided output stream
+            # if possible
+            if output is None:
+                output = BytesIO()
+            else:
+                # Rationale for the explicit writability check:
+                #  If the output buffer is not readable or not seekable, it's
+                #  about to be replaced with a BytesIO instance, and in that
+                #  case, the write error would only happen *after* the signing
+                #  operations are done. We want to avoid that scenario.
+                if not output.writable():
+                    raise IOError(
+                        "Output buffer is not writable"
+                    )  # pragma: nocover
+                if not output.seekable() or not output.readable():
+                    output = BytesIO()
+
             writer.write(output)
 
         # retcon time: write the proper values of the /ByteRange entry
         #  in the signature object
         eof = output.tell()
-        sig_start, sig_end = self.signature_contents.offsets
+        sig_start, sig_end = self.contents.offsets
         self.byte_range.fill_offsets(output, sig_start, sig_end, eof)
 
         # compute the digests
@@ -215,21 +202,21 @@ class PdfSignedData(generic.DictionaryObject):
             misc.chunked_digest(temp_buffer, output, md, max_read=eof-sig_end)
 
         digest_value = md.digest()
-        signature_cms = yield digest_value
+        cms_data = yield digest_value
 
-        if isinstance(signature_cms, bytes):
-            signature_bytes = signature_cms
+        if isinstance(cms_data, bytes):
+            der_bytes = cms_data
         else:
-            signature_bytes = signature_cms.dump()
-        signature = binascii.hexlify(signature_bytes).upper()
+            der_bytes = cms_data.dump()
+        cms_hex = binascii.hexlify(der_bytes).upper()
 
         # might as well compute this
         bytes_reserved = sig_end - sig_start - 2
-        length = len(signature)
+        length = len(cms_hex)
         if length > bytes_reserved:
             raise SigningError(
-                f"Final signature buffer larger than expected: "
-                f"allocated {bytes_reserved} bytes, but signature required"
+                f"Final CMS buffer larger than expected: "
+                f"allocated {bytes_reserved} bytes, but CMS required "
                 f"{length} bytes."
             )  # pragma: nocover
 
@@ -239,11 +226,48 @@ class PdfSignedData(generic.DictionaryObject):
         # signature contents are NOT supposed to be encrypted.
         # Perhaps this falls under the "strings in encrypted containers"
         # denominator in § 7.6.1?
-        output.write(signature)
+        # Addition: the PDF 2.0 spec *does* spell out that this content
+        # is not to be encrypted.
+        output.write(cms_hex)
 
         output.seek(0)
-        padding = bytes(bytes_reserved // 2 - len(signature_bytes))
-        yield output, signature_bytes + padding
+        padding = bytes(bytes_reserved // 2 - len(der_bytes))
+        yield output, der_bytes + padding
+
+
+class PdfSignedData(PdfByteRangeDigest):
+    """
+    Generic class to model signature dictionaries in a PDF file.
+    See also :class:`.SignatureObject` and :class:`.DocumentTimestamp`.
+
+    :param obj_type:
+        The type of signature object.
+    :param subfilter:
+        See :class:`.SigSeedSubFilter`.
+    :param timestamp:
+        The timestamp to embed into the ``/M`` entry.
+    :param bytes_reserved:
+        The number of bytes to reserve for the signature.
+        Defaults to 16 KiB.
+
+        .. warning::
+            Since the CMS object is written to the output file as a hexadecimal
+            string, you should request **twice** the (estimated) number of bytes
+            in the DER-encoded version of the CMS object.
+    """
+
+    def __init__(self, obj_type,
+                 subfilter: SigSeedSubFilter = DEFAULT_SIG_SUBFILTER,
+                 timestamp: datetime = None, bytes_reserved=None):
+        super().__init__(bytes_reserved=bytes_reserved)
+        self.update({
+            pdf_name('/Type'): obj_type,
+            pdf_name('/Filter'): pdf_name('/Adobe.PPKLite'),
+            pdf_name('/SubFilter'): subfilter.value,
+        })
+
+        if timestamp is not None:
+            self[pdf_name('/M')] = pdf_date(timestamp)
 
 
 class SignatureObject(PdfSignedData):
@@ -337,7 +361,7 @@ class Signer:
     about trust.
     """
 
-    signature_mechanism: SignedDigestAlgorithm
+    signature_mechanism: SignedDigestAlgorithm = None
     """
     The (cryptographic) signature mechanism to use.
     """
@@ -463,20 +487,26 @@ class Signer:
                 'adobe_revocation_info_archival', revinfo
             )
 
-        return None
-
-    def signed_attrs(self, data_digest: bytes, timestamp: datetime = None,
+    def signed_attrs(self, data_digest: bytes, digest_algorithm: str,
+                     timestamp: datetime = None,
                      revocation_info=None, use_pades=False):
         """
+        .. versionchanged:: 0.4.0
+            Added positional ``digest_algorithm`` parameter _(breaking change)_.
+
         Format the signed attributes for a CMS signature.
 
         :param data_digest:
             Raw digest of the data to be signed.
+        :param digest_algorithm:
+            .. versionadded:: 0.4.0
+
+            Name of the digest algorithm used to compute the digest.
         :param timestamp:
             Current timestamp (ignored when ``use_pades`` is ``True``).
         :param revocation_info:
             Revocation information to embed; this should be the output
-            of a call to :meth:`.Signer.format_revinfo`
+            of a call to :meth:`.Signer.format_revinfo` or ``None``
             (ignored when ``use_pades`` is ``True``).
         :param use_pades:
             Respect PAdES requirements.
@@ -509,8 +539,67 @@ class Signer:
             if revocation_info:
                 attrs.append(revocation_info)
 
+            # TODO not sure if PAdES allows this, need to check.
+            #  It *should*, but perhaps the version of CMS it is based on is too
+            #  old, or it might not allow undefined signed attributes.
+            # In the meantime, we only add this attribute to non-PAdES sigs
+            algid_protection = cms.CMSAlgorithmProtection({
+                'digest_algorithm': algos.DigestAlgorithm(
+                    {'algorithm': digest_algorithm}
+                ),
+                'signature_algorithm':
+                    self.get_signature_mechanism(digest_algorithm)
+            })
+            attrs.append(
+                simple_cms_attribute(
+                    'cms_algorithm_protection', algid_protection
+                )
+            )
+
         return cms.CMSAttributes(attrs)
 
+    def unsigned_attrs(self, digest_algorithm, signature: bytes,
+                       timestamper=None, dry_run=False) \
+            -> Optional[cms.CMSAttributes]:
+        """
+        Compute the unsigned attributes to embed into the CMS object.
+        This function is called after signing the hash of the signed attributes
+        (see :meth:`signed_attrs`).
+
+        By default, this method only handles timestamp requests, but other
+        functionality may be added by subclasses
+
+        If this method returns ``None``, no unsigned attributes will be
+        embedded.
+
+        :param digest_algorithm:
+            Digest algorithm used to hash the signed attributes.
+        :param signature:
+            Signature of the signed attribute hash.
+        :param timestamper:
+            Timestamp supplier to use.
+        :param dry_run:
+            Flag indicating "dry run" mode. If ``True``, only the approximate
+            size of the output matters, so cryptographic
+            operations can be replaced by placeholders.
+        :return:
+            The unsigned attributes to add, or ``None``.
+        """
+
+        if timestamper is not None:
+            # the timestamp server needs to cross-sign our signature
+            md = getattr(hashlib, digest_algorithm)()
+            md.update(signature)
+            if dry_run:
+                ts_token = timestamper.dummy_response(digest_algorithm)
+            else:
+                ts_token = timestamper.timestamp(
+                    md.digest(), digest_algorithm
+                )
+            return cms.CMSAttributes(
+                [simple_cms_attribute('signature_time_stamp_token', ts_token)]
+            )
+
     def signer_info(self, digest_algorithm: str, signed_attrs, signature):
         """
         Format the ``SignerInfo`` entry for a CMS signature.
@@ -592,18 +681,17 @@ class Signer:
             An :class:`~.asn1crypto.cms.ContentInfo` object.
         """
 
+        digest_algorithm = digest_algorithm.lower()
         # the piece of data we'll actually sign is a DER-encoded version of the
         # signed attributes of our message
         signed_attrs = self.signed_attrs(
-            data_digest, timestamp, revocation_info=revocation_info,
-            use_pades=use_pades
+            data_digest, digest_algorithm, timestamp,
+            revocation_info=revocation_info, use_pades=use_pades
         )
 
-        # TODO decouple the document hashing MD from the CMS-internal MD
-        #  it's probably a good idea to allow digest_algorithm to be None
-        #  here if it's implied by the signature mechanism, but care is needed
-        #  to integrate it correctly with the signer and validator
-        digest_algorithm = digest_algorithm.lower()
+        digest_algorithm_obj = algos.DigestAlgorithm(
+            {'algorithm': digest_algorithm}
+        )
         implied_hash_algo = None
         try:
             if self.signature_mechanism is not None:
@@ -625,23 +713,12 @@ class Signer:
 
         sig_info = self.signer_info(digest_algorithm, signed_attrs, signature)
 
-        if timestamper is not None:
-            # the timestamp server needs to cross-sign our signature
-            md = getattr(hashlib, digest_algorithm)()
-            md.update(signature)
-            if dry_run:
-                ts_token = timestamper.dummy_response(digest_algorithm)
-            else:
-                ts_token = timestamper.timestamp(
-                    md.digest(), digest_algorithm
-                )
-            sig_info['unsigned_attrs'] = cms.CMSAttributes(
-                [simple_cms_attribute('signature_time_stamp_token', ts_token)]
-            )
-
-        digest_algorithm_obj = algos.DigestAlgorithm(
-            {'algorithm': digest_algorithm}
+        unsigned_attrs = self.unsigned_attrs(
+            digest_algorithm, signature, timestamper=timestamper,
+            dry_run=dry_run
         )
+        if unsigned_attrs is not None:
+            sig_info['unsigned_attrs'] = unsigned_attrs
 
         # do not add the TS certs at this point
         certs = set(self.cert_registry)
@@ -877,7 +954,7 @@ class SimpleSigner(Signer):
         try:
             return set(load_certs_from_pemder(ca_chain_files))
         except (IOError, ValueError) as e:  # pragma: nocover
-            logger.error('Could not load CA chain', e)
+            logger.error('Could not load CA chain', exc_info=e)
             return None
 
     @classmethod
@@ -909,7 +986,7 @@ class SimpleSigner(Signer):
             with open(pfx_file, 'rb') as f:
                 pfx_bytes = f.read()
         except IOError as e:  # pragma: nocover
-            logger.error(f'Could not open PKCS#12 file {pfx_file}.', e)
+            logger.error(f'Could not open PKCS#12 file {pfx_file}.', exc_info=e)
             return None
 
         ca_chain = cls._load_ca_chain(ca_chain_files) \
@@ -964,7 +1041,7 @@ class SimpleSigner(Signer):
                     f.read()
                 )
         except (IOError, ValueError) as e:  # pragma: nocover
-            logger.error('Could not load cryptographic material', e)
+            logger.error('Could not load cryptographic material', exc_info=e)
             return None
 
         ca_chain = cls._load_ca_chain(ca_chain_files) if ca_chain_files else []
@@ -1019,7 +1096,7 @@ SIG_DETAILS_DEFAULT_TEMPLATE = (
 )
 
 
-def sign_pdf(pdf_out: IncrementalPdfFileWriter,
+def sign_pdf(pdf_out: BasePdfFileWriter,
              signature_meta: PdfSignatureMetadata, signer: Signer,
              timestamper: TimeStamper = None,
              new_field_spec: Optional[SigFieldSpec] = None,
@@ -1057,9 +1134,6 @@ def sign_pdf(pdf_out: IncrementalPdfFileWriter,
         Write the output to the specified output stream.
         If ``None``, write to a new :class:`.BytesIO` object.
         Default is ``None``.
-
-        .. warning::
-            The output stream must also support reading and seeking.
     :return:
         The output stream containing the signed output.
     """
@@ -1082,7 +1156,8 @@ def sign_pdf(pdf_out: IncrementalPdfFileWriter,
 
 # Wrapper around _prepare_sig_fields with some error reporting
 
-def _get_or_create_sigfield(field_name, pdf_out, existing_fields_only,
+def _get_or_create_sigfield(field_name, pdf_out: BasePdfFileWriter,
+                            existing_fields_only,
                             new_field_spec: Optional[SigFieldSpec] = None):
     root = pdf_out.root
     if field_name is None:
@@ -1097,9 +1172,7 @@ def _get_or_create_sigfield(field_name, pdf_out, existing_fields_only,
         # irrelevant in this special case, so we might as well short circuit
         # things.
         field_created = False
-        empty_fields = enumerate_sig_fields(
-            pdf_out.prev, filled_status=False
-        )
+        empty_fields = enumerate_sig_fields(pdf_out, filled_status=False)
         try:
             found_field_name, _, sig_field_ref = next(empty_fields)
         except StopIteration:
@@ -1342,9 +1415,6 @@ class SigIOSetup:
     """
     Write the output to the specified output stream. If ``None``, write to a 
     new :class:`.BytesIO` object. Default is ``None``.
-
-    .. warning::
-        The output stream must also support reading and seeking.
     """
 
 
@@ -1364,7 +1434,7 @@ class PdfCMSEmbedder:
     def __init__(self, new_field_spec: Optional[SigFieldSpec] = None):
         self.new_field_spec = new_field_spec
 
-    def write_cms(self, field_name: str, writer: IncrementalPdfFileWriter,
+    def write_cms(self, field_name: str, writer: BasePdfFileWriter,
                   existing_fields_only=False):
         """
         This method returns a generator coroutine that controls the process
@@ -1470,12 +1540,28 @@ class PdfCMSEmbedder:
         assert isinstance(sig_io, SigIOSetup)
 
         # pass control to the sig object's write_signature coroutine
-        yield from sig_obj.write_signature(
+        yield from sig_obj.fill(
             writer, sig_io.md_algorithm, in_place=sig_io.in_place,
             output=sig_io.output, chunk_size=sig_io.chunk_size
         )
 
 
+def _finalise_output(orig_output, returned_output):
+
+    # The internal API transparently replaces non-readable/seekable
+    # buffers with BytesIO for signing operations, but we don't want to
+    # expose that to the public API user.
+
+    if orig_output is not None and orig_output is not returned_output:
+        # original output is a write-only buffer
+        assert isinstance(returned_output, BytesIO)
+        raw_buf = returned_output.getbuffer()
+        orig_output.write(raw_buf)
+        raw_buf.release()
+        return orig_output
+    return returned_output
+
+
 class PdfTimeStamper:
     """
     Class to encapsulate the process of appending document timestamps to
@@ -1533,9 +1619,6 @@ class PdfTimeStamper:
             Write the output to the specified output stream.
             If ``None``, write to a new :class:`.BytesIO` object.
             Default is ``None``.
-
-            .. warning::
-                The output stream must also support reading and seeking.
         :param in_place:
             Sign the original input stream in-place.
             This parameter overrides ``output``.
@@ -1579,15 +1662,17 @@ class PdfTimeStamper:
         )
         true_digest = cms_writer.send(sig_io)
         timestamp_cms = timestamper.timestamp(true_digest, md_algorithm)
-        output, sig_contents = cms_writer.send(timestamp_cms)
+        res_output, sig_contents = cms_writer.send(timestamp_cms)
 
         # update the DSS
         from pyhanko.sign import validation
         validation.DocumentSecurityStore.add_dss(
-            output_stream=output, sig_contents=sig_contents,
+            output_stream=res_output, sig_contents=sig_contents,
             paths=validation_paths, validation_context=validation_context
         )
 
+        output = _finalise_output(output, res_output)
+
         return output
 
     def update_archival_timestamp_chain(self, reader: PdfFileReader,
@@ -1810,10 +1895,7 @@ class PdfSigner(PdfTimeStamper):
             )
         if cd.permission == MDPPerm.NO_CHANGES:
             raise SigningError("Author signature forbids all changes")
-        author_sig_md = cd.author_sig.get('/DigestAlgorithm', None)
-        if author_sig_md is not None:
-            return author_sig_md
-        return None
+        return cd.digest_method
 
     def _enforce_seed_value_constraints(self, sig_field, validation_path) \
             -> Optional[SigSeedValueSpec]:
@@ -1937,7 +2019,7 @@ class PdfSigner(PdfTimeStamper):
         # LOCK_DOCUMENT is only enforced later
         return sv_spec
 
-    def sign_pdf(self, pdf_out: IncrementalPdfFileWriter,
+    def sign_pdf(self, pdf_out: BasePdfFileWriter,
                  existing_fields_only=False, bytes_reserved=None, *,
                  appearance_text_params=None, in_place=False,
                  output=None, chunk_size=4096):
@@ -1945,7 +2027,8 @@ class PdfSigner(PdfTimeStamper):
         Sign a PDF file using the provided output writer.
 
         :param pdf_out:
-            An :class:`.IncrementalPdfFileWriter` containing the data to sign.
+            A PDF file writer (usually an :class:`.IncrementalPdfFileWriter`)
+            containing the data to sign.
         :param existing_fields_only:
             If ``True``, never create a new empty signature field to contain
             the signature.
@@ -1961,9 +2044,6 @@ class PdfSigner(PdfTimeStamper):
             Write the output to the specified output stream.
             If ``None``, write to a new :class:`.BytesIO` object.
             Default is ``None``.
-
-            .. warning::
-                The output stream must also support reading and seeking.
         :param in_place:
             Sign the original input stream in-place.
             This parameter overrides ``output``.
@@ -2024,9 +2104,11 @@ class PdfSigner(PdfTimeStamper):
             sig_field, signer_cert_validation_path
         )
 
-        author_sig_md_algorithm = self._enforce_certification_constraints(
-            pdf_out.prev
-        )
+        author_sig_md_algorithm = None
+        if isinstance(pdf_out, IncrementalPdfFileWriter):
+            author_sig_md_algorithm = self._enforce_certification_constraints(
+                pdf_out.prev
+            )
         # priority order for the message digest algorithm
         #  (1) If signature_meta specifies a message digest algorithm, use it
         #      (it has been cleared by the SV dictionary checker already)
@@ -2135,22 +2217,25 @@ class PdfSigner(PdfTimeStamper):
             revocation_info=revinfo, timestamper=timestamper
         )
         # ... and feed it to the CMS writer
-        output, sig_contents = cms_writer.send(signature_cms)
+        res_output, sig_contents = cms_writer.send(signature_cms)
 
         if use_pades and signature_meta.embed_validation_info:
             from pyhanko.sign import validation
             validation.DocumentSecurityStore.add_dss(
-                output_stream=output, sig_contents=sig_contents,
+                output_stream=res_output, sig_contents=sig_contents,
                 paths=validation_paths, validation_context=validation_context
             )
 
             if timestamper is not None and signature_meta.use_pades_lta:
                 # append an LTV document timestamp
-                w = IncrementalPdfFileWriter(output)
+                w = IncrementalPdfFileWriter(res_output)
                 self.timestamp_pdf(
                     w, md_algorithm, validation_context,
                     validation_paths=ts_validation_paths, in_place=True,
                     timestamper=timestamper, chunk_size=chunk_size
                 )
 
+        # we put the finalisation step after the DSS manipulations, since
+        # otherwise we'd also run into issues with non-seekable output buffers
+        output = _finalise_output(output, res_output)
         return output
