@@ -5,7 +5,7 @@ from collections import namedtuple
 from dataclasses import dataclass, field as data_field
 from datetime import datetime
 from enum import Enum, unique
-from typing import TypeVar, Type, Optional, Union
+from typing import TypeVar, Type, Optional, Union, Tuple
 
 from asn1crypto import (
     cms, tsp, ocsp as asn1_ocsp, pdf as asn1_pdf, crl as asn1_crl, x509,
@@ -46,7 +46,7 @@ __all__ = [
     'read_certification_data', 'validate_pdf_ltv_signature',
     'validate_pdf_signature', 'validate_cms_signature',
     'ValidationInfoReadingError', 'SignatureValidationError',
-    'SigSeedValueValidationError'
+    'SigSeedValueValidationError', 'validate_sig_integrity'
 ]
 
 logger = logging.getLogger(__name__)
@@ -147,13 +147,128 @@ def _validate_raw(signature: bytes, signed_blob: bytes, cert: x509.Certificate,
     )
 
 
+def validate_sig_integrity(signer_info: cms.SignedData,
+                           cert: x509.Certificate,
+                           expected_content_type: str,
+                           actual_digest: bytes) -> Tuple[bool, bool]:
+    """
+    Validate the integrity of a signature for a particular signerInfo object
+    inside a CMS signed data container.
+
+    .. warning::
+        This function does not do any trust checks, and is considered
+        "dangerous" API because it is easy to misuse.
+
+    .. warning::
+        This function currently does not deal with the case where signed
+        attributes are absent.
+
+    :param signer_info:
+        A :class:`cms.SignerInfo` object.
+    :param cert:
+        The signer's certificate.
+
+        .. note::
+            This function will not attempt to extract certificates from
+            the signed data.
+    :param expected_content_type:
+        The expected value for the content type attribute (as a Python string,
+        see :class:`cms.ContentType`).
+    :param actual_digest:
+        The actual digest to be matched to the message digest attribute.
+    :return:
+        A tuple of two booleans. The first indicates whether the provided
+        digest matches the value in the signed attributes.
+        The second indicates whether the signature of the digest is valid.
+    """
+
+    signature_algorithm: cms.SignedDigestAlgorithm = \
+        signer_info['signature_algorithm']
+    digest_algorithm_obj = signer_info['digest_algorithm']
+    md_algorithm = digest_algorithm_obj['algorithm'].native
+    signature = signer_info['signature'].native
+
+    # signed_attrs comes with some context-specific tagging
+    # because it's an implicit field. This breaks validation
+    signed_attrs = signer_info['signed_attrs'].untag()
+    # TODO if there are no signed_attrs, we should validate the signature
+    #  against actual_digest. Find some real-world exmples to test this
+    #  Also, signed_attrs is mandatory if content_type is not id-data
+
+    # check the CMSAlgorithmProtection attr, if present
+    try:
+        cms_algid_protection, = find_cms_attribute(
+            signed_attrs, 'cms_algorithm_protection'
+        )
+        signed_digest_algorithm = \
+            cms_algid_protection['digest_algorithm'].native
+        if signed_digest_algorithm != digest_algorithm_obj.native:
+            raise SignatureValidationError(
+                "Digest algorithm does not match CMS algorithm protection "
+                "attribute."
+            )
+        signed_sig_algorithm = \
+            cms_algid_protection['signature_algorithm'].native
+        if signed_sig_algorithm is None:
+            raise SignatureValidationError(
+                "CMS algorithm protection attribute not valid for signed data"
+            )
+        elif signed_sig_algorithm != signature_algorithm.native:
+            raise SignatureValidationError(
+                "Signature mechanism does not match CMS algorithm "
+                "protection attribute."
+            )
+    except KeyError:
+        pass
+    except SignatureValidationError:
+        raise
+    except ValueError:
+        raise SignatureValidationError(
+            'Multiple CMS protection attributes present'
+        )
+
+    signed_blob = signed_attrs.dump(force=True)
+    try:
+        content_type, = find_cms_attribute(signed_attrs, 'content_type')
+        content_type = content_type.native
+        if content_type != expected_content_type:
+            raise SignatureValidationError(
+                'Content type did not match expected value'
+            )
+    except (KeyError, ValueError):
+        raise SignatureValidationError(
+            'Content type not found in signature, or multiple content-type '
+            'attributes present.'
+        )
+
+    try:
+        embedded_digest, = find_cms_attribute(signed_attrs, 'message_digest')
+        embedded_digest = embedded_digest.native
+    except (KeyError, ValueError):
+        raise SignatureValidationError(
+            'Message digest not found in signature, or multiple message '
+            'digest attributes present.'
+        )
+    intact = actual_digest == embedded_digest
+
+    try:
+        _validate_raw(
+            signature, signed_blob, cert, signature_algorithm, md_algorithm
+        )
+        valid = True
+    except SignatureError:
+        valid = False
+
+    return intact, valid
+
+
 def _validate_cms_signature(signed_data: cms.SignedData,
                             status_cls: Type[StatusType] = SignatureStatus,
                             raw_digest: bytes = None,
                             validation_context: ValidationContext = None,
                             status_kwargs: dict = None,
                             key_usage_settings: KeyUsageConstraints = None,
-                            externally_invalid=False):
+                            encap_data_invalid=False):
     """
     Validate CMS and PKCS#7 signatures.
     """
@@ -161,44 +276,29 @@ def _validate_cms_signature(signed_data: cms.SignedData,
 
     signature_algorithm: cms.SignedDigestAlgorithm = \
         signer_info['signature_algorithm']
-    mechanism = signature_algorithm['algorithm'].native.lower()
-    md_algorithm = \
-        signer_info['digest_algorithm']['algorithm'].native.lower()
-    signature = signer_info['signature'].native
-    # signed_attrs comes with some context-specific tagging
-    # because it's an implicit field. This breaks validation
-    signed_attrs = signer_info['signed_attrs'].untag()
+    mechanism = signature_algorithm['algorithm'].native
+    md_algorithm = signer_info['digest_algorithm']['algorithm'].native
 
-    # TODO What to do if signed_attrs is absent?
-    # I guess I'll wait until someone complains that a valid signature
-    # isn't being validated correctly
+    expected_content_type = 'data'
     if raw_digest is None:
         # this means that there should be encapsulated data
-        # TODO Carefully read § 5.2.1 in RFC 5652, and compare with
-        #  the implementation in asn1crypto.
-        raw = signed_data['encap_content_info']['content'].parsed.dump()
+        eci = signed_data['encap_content_info']
+        expected_content_type = eci['content_type'].native
+
+        raw = eci['content'].parsed.dump()
         raw_digest = getattr(hashlib, md_algorithm)(raw).digest()
 
-    signed_blob = signed_attrs.dump(force=True)
-    try:
-        embedded_digest = find_cms_attribute(signed_attrs, 'message_digest')
-    except KeyError:
-        raise SignatureValidationError('Message digest not found in signature')
-    intact = raw_digest == embedded_digest[0].native
-
-    valid = False
-    if intact:
-        try:
-            _validate_raw(
-                signature, signed_blob, cert, signature_algorithm, md_algorithm
-            )
-            valid = True
-        except SignatureError:
-            valid = False
+    # first, do the cryptographic identity checks
+    intact, valid = validate_sig_integrity(
+        signer_info, cert, expected_content_type=expected_content_type,
+        actual_digest=raw_digest,
+    )
 
     # if the signature is invalid for some external reason, this flag is set
     #  (e.g. when the thing being signed is itself wrong)
-    valid &= not externally_invalid
+    valid &= not encap_data_invalid and intact
+
+    # next, validate trust
     trusted = revoked = False
     path = None
     if valid:
@@ -226,7 +326,7 @@ def validate_cms_signature(signed_data: cms.SignedData,
                            validation_context: ValidationContext = None,
                            status_kwargs: dict = None,
                            key_usage_settings: KeyUsageConstraints = None,
-                           externally_invalid=False):
+                           encap_data_invalid=False):
     """
     Validate a detached CMS signature (i.e. a ``SignedData`` object).
 
@@ -244,9 +344,11 @@ def validate_cms_signature(signed_data: cms.SignedData,
     :param key_usage_settings:
         A :class:`.KeyUsageConstraints` object specifying which key usage
         extensions must or must not be present in the signer's certificate.
-    :param externally_invalid:
-        If ``True``, there is an external reason why the signature cannot
-        be valid, but the remaining validation logic still has to be run.
+    :param encap_data_invalid:
+        If ``True``, the encapsulated data inside the CMS is invalid,
+        but the remaining validation logic still has to be run (e.g. a
+        timestamp token, which requires validation of the embedded message
+        imprint).
 
         This option is considered internal API, the semantics of which may
         change without notice in the future.
@@ -255,7 +357,7 @@ def validate_cms_signature(signed_data: cms.SignedData,
     """
     status_kwargs = _validate_cms_signature(
         signed_data, status_cls, raw_digest, validation_context,
-        status_kwargs, key_usage_settings, externally_invalid
+        status_kwargs, key_usage_settings, encap_data_invalid
     )
     return status_cls(**status_kwargs)
 
@@ -355,12 +457,14 @@ class PdfSignatureStatus(SignatureStatus):
         an instance of :class:`.ModificationLevel` otherwise.
         """
 
+        coverage = self.coverage
         if self.diff_result is None:
-            if self.coverage == SignatureCoverageLevel.ENTIRE_REVISION:
+            if coverage == SignatureCoverageLevel.ENTIRE_REVISION:
                 # in this case, we can't know without the diff analysis result
                 return None
             return (
-                ModificationLevel.NONE if SignatureCoverageLevel.ENTIRE_FILE
+                ModificationLevel.NONE
+                if coverage == SignatureCoverageLevel.ENTIRE_FILE
                 else ModificationLevel.OTHER
             )
         elif isinstance(self.diff_result, DiffResult):
@@ -412,7 +516,10 @@ class PdfSignatureStatus(SignatureStatus):
         if self.coverage == SignatureCoverageLevel.ENTIRE_FILE:
             yield 'UNTOUCHED'
         elif self.coverage == SignatureCoverageLevel.ENTIRE_REVISION:
-            yield 'EXTENDED_WITH_' + self.modification_level.name
+            if self.modification_level is not None:
+                yield 'EXTENDED_WITH_' + self.modification_level.name
+            else:
+                yield 'EXTENDED'
         else:
             yield 'NONSTANDARD_COVERAGE'
         if self.docmdp_ok:
@@ -585,7 +692,7 @@ class EmbeddedPdfSignature:
     """
 
     def __init__(self, reader: PdfFileReader,
-                 sig_field: generic.DictionaryObject):
+                 sig_field: generic.DictionaryObject, fq_name: str):
         self.reader = reader
         if isinstance(sig_field, generic.IndirectObject):
             sig_field = sig_field.get_object()
@@ -627,7 +734,17 @@ class EmbeddedPdfSignature:
         self.md_algorithm = digest_algo['algorithm'].native.lower()
 
         # grab the revision to which the signature applies
-        self.signed_revision = self.reader.xrefs.get_introducing_revision(
+        # NOTE: We're using get_last_change here as opposed to
+        # get_introducing_revision. The distinction won't be relevant in most
+        # legitimate use cases, but get_last_change is more likely to be correct
+        # in cases where the signature obj was created by overriding an existing
+        # object (which is weird, but technically possible, I guess).
+        # Important note: the coverage checker will validate whether the
+        # xref table for that revision is actually covered by the signature,
+        # and raise the alarm if that's not the case.
+        # Therefore shenanigans with updating signature objects will be detected
+        # even before the diff checker runs.
+        self.signed_revision = self.reader.xrefs.get_last_change(
             sig_object_ref.reference
         )
         self.coverage = None
@@ -639,6 +756,7 @@ class EmbeddedPdfSignature:
 
         self.diff_result = None
         self._integrity_checked = False
+        self.fq_name = fq_name
 
     @property
     def field_name(self):
@@ -646,7 +764,7 @@ class EmbeddedPdfSignature:
         :return:
             Name of the signature field.
         """
-        return self.sig_field['/T']
+        return self.fq_name
 
     # TODO also parse the signature object's /M entry
     @property
@@ -1269,15 +1387,15 @@ def _validate_timestamp(tst_signed_data, validation_context,
             f"Timestamp token imprint is {tst_imprint.hex()}, but expected "
             f"{expected_tst_imprint.hex()}."
         )
-        externally_invalid = True
+        encap_data_invalid = True
     else:
-        externally_invalid = False
+        encap_data_invalid = False
     timestamp = tst_info['gen_time'].native
     timestamp_status: TimestampSignatureStatus = validate_cms_signature(
         tst_signed_data, status_cls=TimestampSignatureStatus,
         validation_context=validation_context,
         status_kwargs={'timestamp': timestamp},
-        externally_invalid=externally_invalid
+        encap_data_invalid=encap_data_invalid
     )
     return timestamp_status
 
@@ -1331,6 +1449,8 @@ def _establish_timestamp_trust_lta(reader, bootstrap_validation_context,
 
 # TODO verify formal PAdES requirements for timestamps
 # TODO verify other formal PAdES requirements (coverage, etc.)
+# TODO signature/verification policy-based validation! (PAdES-EPES-* etc)
+#  (this is a different beast, though)
 def validate_pdf_ltv_signature(embedded_sig: EmbeddedPdfSignature,
                                validation_type: RevocationInfoValidationType,
                                validation_context_kwargs=None,
@@ -1505,7 +1625,9 @@ def apply_adobe_revocation_info(signer_info: cms.SignerInfo,
     )
 
 
-DocMDPInfo = namedtuple('DocMDPInfo', ['permission', 'author_sig'])
+DocMDPInfo = namedtuple(
+    'DocMDPInfo', ['permission', 'digest_method', 'author_sig']
+)
 """
 Encodes certification information for a signed document, consisting of a 
 reference to the author signature, together with the associated DocMDP policy.
@@ -1527,8 +1649,14 @@ def read_certification_data(reader: PdfFileReader) -> Optional[DocMDPInfo]:
         return
 
     perm = _extract_docmdp_for_sig(certification_sig)
+    ref = _extract_reference_dict(certification_sig, '/DocMDP')
+    md = None
+    if ref is not None:
+        md = misc.get_and_apply(
+            ref, '/DigestMethod', lambda x: x[1:].lower()
+        )
 
-    return DocMDPInfo(perm, certification_sig)
+    return DocMDPInfo(perm, md, certification_sig)
 
 
 @dataclass
@@ -1558,12 +1686,6 @@ class VRI:
     Relevant CRLs.
     """
 
-    def __iadd__(self, other):
-        self.certs.update(other.certs)
-        self.crls.update(other.crls)
-        self.ocsps.update(other.ocsps)
-        return self
-
     def as_pdf_object(self) -> generic.DictionaryObject:
         """
         :return:
@@ -1672,45 +1794,51 @@ class DocumentSecurityStore:
         ident = hashlib.sha1(contents).digest().hex().upper()
         return pdf_name('/' + ident)
 
-    def register_vri(self, identifier, paths, validation_context):
+    def register_vri(self, identifier, *, certs=(), ocsps=(), crls=()):
         """
         Register validation information for a set of signing certificates
         associated with a particular signature.
 
         :param identifier:
-            Identifier of the signature object (see `sig_content_identifier`)
-        :param paths:
-            Validation paths to add.
-        :param validation_context:
-            Validation context to source CRLs and OCSP responses from.
+            Identifier of the signature object (see `sig_content_identifier`).
+            If ``None``, only embed the data into the DSS without associating
+            it with any VRI.
+        :param certs:
+            Certificates to add.
+        :param ocsps:
+            OCSP responses to add.
+        :param crls:
+            CRLs to add.
         """
 
         if self.writer is None:
             raise TypeError('This DSS does not support updates.')
 
-        # embed any hardcoded ocsp responses and CRLs, if applicable
-        ocsps = set(
-            self._cms_objects_to_streams(
-                validation_context.ocsps, self._ocsps_seen, self.ocsps
+        ocsps = list(ocsps)
+        crls = list(crls)
+
+        ocsp_refs = set()
+        crl_refs = set()
+        cert_refs = {self._embed_cert(cert) for cert in certs}
+        if ocsps:
+            ocsp_refs = set(
+                self._cms_objects_to_streams(
+                    ocsps, self._ocsps_seen, self.ocsps
+                )
             )
-        )
-        crls = set(
-            self._cms_objects_to_streams(
-                validation_context.crls, self._crls_seen, self.crls
+        if crls:
+            crl_refs = set(
+                self._cms_objects_to_streams(crls, self._crls_seen, self.crls)
             )
-        )
-        path: ValidationPath
         # TODO while somewhat less common, CRL signing can also be delegated
         #  we should take that into account
-        cert_refs = set(self._embed_certs_from_ocsp(validation_context.ocsps))
-        for path in paths:
-            for cert in path:
-                cert_refs.add(self._embed_cert(cert))
+        cert_refs.update(set(self._embed_certs_from_ocsp(ocsps)))
 
-        vri = VRI(certs=cert_refs, ocsps=ocsps, crls=crls)
-        self.vri_entries[identifier] = self.writer.add_object(
-            vri.as_pdf_object()
-        )
+        if identifier is not None:
+            vri = VRI(certs=cert_refs, ocsps=ocsp_refs, crls=crl_refs)
+            self.vri_entries[identifier] = self.writer.add_object(
+                vri.as_pdf_object()
+            )
 
     def as_pdf_object(self):
         """
@@ -1835,16 +1963,29 @@ class DocumentSecurityStore:
         return dss
 
     @classmethod
-    def add_dss(cls, output_stream, sig_contents, paths,
-                validation_context):
+    def add_dss(cls, output_stream, sig_contents, *, certs=None,
+                ocsps=None, crls=None, paths=None, validation_context=None):
         """
-        Add or update a DSS, and add the new information to a specific VRI.
-        This will be done as an incremental update.
+        Add or update a DSS, and optionally associate the new information with a
+        VRI entry tied to a signature object.
+
+        The result is applied to the output stream as an incremental update.
+
+        You can either specify the CMS objects to include directly, or
+        pass them in as output from `certvalidator`.
 
         :param output_stream:
             Output stream to write to.
         :param sig_contents:
-            Contents of the new signature (used to compute the VRI hash)
+            Contents of the new signature (used to compute the VRI hash).
+            If ``None``, the information will not be added to any VRI
+            dictionary.
+        :param certs:
+            Certificates to include in the VRI entry.
+        :param ocsps:
+            OCSP responses to include in the VRI entry.
+        :param crls:
+            CRLs to include in the VRI entry.
         :param paths:
             Validation paths that have been established, and need to be added
             to the DSS.
@@ -1860,9 +2001,32 @@ class DocumentSecurityStore:
             created = True
             dss = cls(writer=writer)
 
-        identifier = DocumentSecurityStore.sig_content_identifier(sig_contents)
+        if sig_contents is not None:
+            identifier = DocumentSecurityStore.sig_content_identifier(
+                sig_contents
+            )
+        else:
+            identifier = None
 
-        dss.register_vri(identifier, paths, validation_context)
+        def _certs():
+            yield from certs or ()
+            path: ValidationPath
+            for path in (paths or ()):
+                yield from path
+
+        def _ocsps():
+            yield from ocsps or ()
+            if validation_context is not None:
+                yield from validation_context.ocsps
+
+        def _crls():
+            yield from crls or ()
+            if validation_context is not None:
+                yield from validation_context.crls
+
+        dss.register_vri(
+            identifier, certs=_certs(), ocsps=_ocsps(), crls=_crls()
+        )
         dss_dict = dss.as_pdf_object()
         # if we're updating the DSS, this is all we need to do.
         # if we're adding a fresh DSS, we need to register it.
